/*
 * Copyright (c) 田梓萱[小草林] 2021-2024.
 * All Rights Reserved.
 * All codes are protected by China's regulations on the protection of computer software, and infringement must be investigated.
 * 版权所有 (c) 田梓萱[小草林] 2021-2024.
 * 所有代码均受中国《计算机软件保护条例》保护，侵权必究.
 */
package io.fastkv

import io.fastkv.interfaces.FastEncoder

/**
 * struct of primary type (boolean/int/float/long/double): [type|keyLen|keyData|value]
 *
 * struct of variable type (string/array/object): [type|keyLen|keyData|valueLen|value]
 *
 * type: 1 byte, keyLen: 1 byte, valueLen: 2 bytes.
 */
internal class Container {
    internal abstract class BaseContainer {
        // The offset record the start of value, not the start of key-value.
        @JvmField var offset = 0
        abstract val type: Byte

        abstract fun equalTo(other: BaseContainer): Boolean
    }

    internal class BooleanContainer(offset: Int, value: Boolean) : BaseContainer() {
        @JvmField var value: Boolean

        init {
            this.offset = offset
            this.value = value
        }

        override val type: Byte
            get() = DataType.BOOLEAN

        override fun equalTo(other: BaseContainer): Boolean {
            return other.type == DataType.BOOLEAN && (other as BooleanContainer).value == value
        }
    }

    internal class IntContainer(offset: Int, value: Int) : BaseContainer() {
        @JvmField var value: Int

        init {
            this.offset = offset
            this.value = value
        }

        override val type: Byte
            get() = DataType.INT

        override fun equalTo(other: BaseContainer): Boolean {
            return other.type == DataType.INT && (other as IntContainer).value == value
        }
    }

    internal class FloatContainer(offset: Int, value: Float) : BaseContainer() {
        @JvmField var value: Float

        init {
            this.offset = offset
            this.value = value
        }

        override val type: Byte
            get() = DataType.FLOAT

        override fun equalTo(other: BaseContainer): Boolean {
            return other.type == DataType.FLOAT && (other as FloatContainer).value == value
        }
    }

    internal class LongContainer(offset: Int, value: Long) : BaseContainer() {
        @JvmField var value: Long

        init {
            this.offset = offset
            this.value = value
        }

        override val type: Byte
            get() = DataType.LONG

        override fun equalTo(other: BaseContainer): Boolean {
            return other.type == DataType.LONG && (other as LongContainer).value == value
        }
    }

    internal class DoubleContainer(offset: Int, value: Double) : BaseContainer() {
        @JvmField var value: Double

        init {
            this.offset = offset
            this.value = value
        }

        override val type: Byte
            get() = DataType.DOUBLE

        override fun equalTo(other: BaseContainer): Boolean {
            return other.type == DataType.DOUBLE && (other as DoubleContainer).value == value
        }
    }

    internal abstract class VarContainer(
        @JvmField var start: Int,
        offset: Int,
        value: Any?,
        size: Int,
        external: Boolean
    ) : BaseContainer() {
        @JvmField var value: Any?
        @JvmField var valueSize: Int
        @JvmField var external: Boolean

        init {
            this.offset = offset
            this.value = value
            valueSize = size
            this.external = external
        }
    }

    internal class StringContainer(
        start: Int,
        offset: Int,
        value: String?,
        size: Int,
        external: Boolean
    ) : VarContainer(start, offset, value, size, external) {
        override val type: Byte
            get() = DataType.STRING

        override fun equalTo(other: BaseContainer): Boolean {
            return other.type == DataType.STRING &&
                (other as StringContainer).external == external &&
                other.value == value
        }
    }

    internal class ArrayContainer(
        start: Int,
        offset: Int,
        value: Any?,
        size: Int,
        external: Boolean
    ) : VarContainer(start, offset, value, size, external) {
        override val type: Byte
            get() = DataType.ARRAY

        override fun equalTo(other: BaseContainer): Boolean {
            if (other.type != DataType.ARRAY) return false
            val otherValue = (other as ArrayContainer).value
            if (value === otherValue) return true
            if (value != null && otherValue != null) {
                if (value is String) {
                    return value == otherValue
                }
                if (value is ByteArray) {
                    return if (otherValue is ByteArray) {
                        (value as ByteArray).contentEquals(otherValue as ByteArray?)
                    } else {
                        false
                    }
                }
            }
            return false
        }
    }

    internal class ObjectContainer(
        start: Int,
        offset: Int,
        value: Any?,
        size: Int,
        external: Boolean
    ) : VarContainer(start, offset, value, size, external) {
        @JvmField var encoder: FastEncoder<*>? = null
        override val type: Byte
            get() = DataType.OBJECT

        override fun equalTo(other: BaseContainer): Boolean {
            return other.type == DataType.OBJECT && (other as ObjectContainer).value == value
        }
    }
}
