/*
 * Copyright (c) 田梓萱[小草林] 2021-2024.
 * All Rights Reserved.
 * All codes are protected by China's regulations on the protection of computer software, and infringement must be investigated.
 * 版权所有 (c) 田梓萱[小草林] 2021-2024.
 * 所有代码均受中国《计算机软件保护条例》保护，侵权必究.
 */
package io.fastkv

import io.fastkv.interfaces.FastCipher
import java.nio.charset.StandardCharsets

internal class FastBuffer @JvmOverloads constructor(var hb: ByteArray, var position: Int = 0) {
    private var charBuf: CharArray? = null

    constructor(capacity: Int) : this(ByteArray(capacity), 0)

    fun get(): Byte {
        return hb[position++]
    }

    fun put(v: Byte) {
        hb[position++] = v
    }

    val short: Short
        get() = (((hb[position++].toInt() and 0xFF)) or ((hb[position++]).toInt() shl 8)).toShort()

    fun putShort(v: Short) {
        hb[position++] = v.toByte()
        hb[position++] = (v.toInt() shr 8).toByte()
    }

    fun getInt(i: Int): Int {
        var i = i
        return (hb[i++].toInt() and
            0xFF or
            (hb[i++].toInt() and 0xFF shl 8) or
            (hb[i++].toInt() and 0xFF shl 16) or
            (hb[i].toInt() shl 24))
    }

    val int: Int
        get() =
            ((((hb[position++].toInt() and 0xFF)) or
                ((hb[position++].toInt() and 0xFF) shl 8) or
                ((hb[position++].toInt() and 0xFF) shl 16) or
                ((hb[position++]).toInt() shl 24)))

    fun getInt(cipher: FastCipher?): Int {
        val value = int
        return cipher?.decrypt(value) ?: value
    }

    fun putInt(v: Int) {
        hb[position++] = v.toByte()
        hb[position++] = (v shr 8).toByte()
        hb[position++] = (v shr 16).toByte()
        hb[position++] = (v shr 24).toByte()
    }

    fun putInt(i: Int, v: Int) {
        var i = i
        hb[i++] = v.toByte()
        hb[i++] = (v shr 8).toByte()
        hb[i++] = (v shr 16).toByte()
        hb[i] = (v shr 24).toByte()
    }

    val varint32: Int
        get() {
            var x = hb[position++].toInt()
            if (x shr 7 == 0) return x
            x = x and 0x7f or (hb[position++].toInt() shl 7)
            if (x shr 14 == 0) return x
            x = x and 0x3fff or (hb[position++].toInt() shl 14)
            if (x shr 21 == 0) return x
            x = x and 0x1fffff or (hb[position++].toInt() shl 21)
            if (x shr 28 == 0) return x
            x = x and 0xfffffff or (hb[position++].toInt() shl 28)
            return x
        }

    private fun putVarint32(i: Int, v: Int): Int {
        var i = i
        var v = v
        while (v and -0x80 != 0) {
            hb[i++] = (v and 0x7f or 0x80).toByte()
            v = v ushr 7
        }
        hb[i++] = v.toByte()
        return i
    }

    fun putVarint32(v: Int) {
        position = putVarint32(position, v)
    }

    fun putLong(i: Int, v: Long) {
        var i = i
        hb[i++] = v.toByte()
        hb[i++] = (v shr 8).toByte()
        hb[i++] = (v shr 16).toByte()
        hb[i++] = (v shr 24).toByte()
        hb[i++] = (v shr 32).toByte()
        hb[i++] = (v shr 40).toByte()
        hb[i++] = (v shr 48).toByte()
        hb[i] = (v shr 56).toByte()
    }

    fun putLong(v: Long) {
        putLong(position, v)
        position += 8
    }

    fun getLong(i: Int): Long {
        var i = i
        return (hb[i++].toLong() and
            0xFFL or
            (hb[i++].toLong() and 0xFFL shl 8) or
            (hb[i++].toLong() and 0xFFL shl 16) or
            (hb[i++].toLong() and 0xFFL shl 24) or
            (hb[i++].toLong() and 0xFFL shl 32) or
            (hb[i++].toLong() and 0xFFL shl 40) or
            (hb[i++].toLong() and 0xFFL shl 48) or
            (hb[i].toLong() shl 56))
    }

    val long: Long
        get() {
            val value = getLong(position)
            position += 8
            return value
        }

    fun getLong(cipher: FastCipher?): Long {
        val value = long
        return cipher?.decrypt(value) ?: value
    }

    fun getFloat(cipher: FastCipher?): Float {
        return java.lang.Float.intBitsToFloat(getInt(cipher))
    }

    fun getDouble(cipher: FastCipher?): Double {
        return java.lang.Double.longBitsToDouble(getLong(cipher))
    }

    private fun getBytes(len: Int): ByteArray {
        val bytes = ByteArray(len)
        System.arraycopy(hb, position, bytes, 0, len)
        position += len
        return bytes
    }

    fun getBytes(cipher: FastCipher?, len: Int): ByteArray? {
        val bytes = getBytes(len)
        return if (cipher != null) cipher.decrypt(bytes) else bytes
    }

    fun putBytes(src: ByteArray) {
        val len = src.size
        if (len > 0) {
            System.arraycopy(src, 0, hb, position, len)
            position += len
        }
    }

    fun getString(cipher: FastCipher?, len: Int): String? {
        if (cipher == null || len <= 0) {
            return getString(len)
        }
        val src = hb.copyOfRange(position, position + len)
        val dst = cipher.decrypt(src)
        val str: String? =
            if (dst == null) {
                null
            } else {
                if (dst.size > MAX_CHAR_LEN) String(dst, StandardCharsets.UTF_8)
                else decodeStr(dst, 0, dst.size)
            }
        position += len
        return str
    }

    fun getString(len: Int): String? {
        return if (len < 0) {
            null
        } else if (len == 0) {
            ""
        } else {
            val str =
                if (len > MAX_CHAR_LEN) String(hb, position, len, StandardCharsets.UTF_8)
                else decodeStr(hb, position, len)
            position += len
            str
        }
    }

    fun putString(str: String?) {
        if (!str.isNullOrEmpty()) {
            encodeStr(str)
        }
    }

    private fun getCharBuf(len: Int): CharArray {
        var buf = charBuf
        if (buf == null) {
            buf =
                if (len <= 256) {
                    CharArray(256)
                } else {
                    CharArray(MAX_CHAR_LEN)
                }
            charBuf = buf
        } else if (buf.size < len) {
            buf = CharArray(MAX_CHAR_LEN)
            charBuf = buf
        }
        return buf
    }

    @Synchronized
    fun decodeStr(src: ByteArray, offset: Int, len: Int): String {
        // Use common buffer to reduce gc.
        val buf = getCharBuf(len)
        var i = offset
        var j = 0
        val limit = offset + len
        while (i < limit) {
            val b1 = src[i++]
            if (b1 > 0) {
                buf[j++] = Char(b1.toUShort())
            } else if (b1 < 0xE0.toByte()) {
                val b2 = src[i++]
                require(!(b1 < 0xC2.toByte() || b2 > 0xBF.toByte())) { INVALID_STRING }
                buf[j++] = (b1.toInt() and 0x1F shl 6 or (b2.toInt() and 0x3F)).toChar()
            } else if (b1 < 0xF0.toByte()) {
                val b2 = src[i++]
                val b3 = src[i++]
                require(
                    !(b1 == 0xE0.toByte() && b2 < 0xA0.toByte() ||
                        b1 == 0xED.toByte() && b2 >= 0xA0.toByte() ||
                        b2 > 0xBF.toByte() ||
                        b3 > 0xBF.toByte())) {
                        INVALID_STRING
                    }
                buf[j++] =
                    (b1.toInt() and
                            0x0F shl
                            12 or
                            (b2.toInt() and 0x3F shl 6) or
                            (b3.toInt() and 0x3F))
                        .toChar()
            } else {
                val b2 = src[i++]
                val b3 = src[i++]
                val b4 = src[i++]
                require(
                    !(b2 > 0xBF.toByte() ||
                        (b1.toInt() shl 28) + (b2 - 0x90.toByte()) shr 30 != 0 ||
                        b3 > 0xBF.toByte() ||
                        b4 > 0xBF.toByte())) {
                        INVALID_STRING
                    }
                val cp =
                    b1.toInt() and
                        0x07 shl
                        18 or
                        (b2.toInt() and 0x3F shl 12) or
                        (b3.toInt() and 0x3F shl 6) or
                        (b4.toInt() and 0x3F)
                buf[j++] = (0xD7C0 + (cp ushr 10)).toChar()
                buf[j++] = (0xDC00 + (cp and 0x3FF)).toChar()
            }
        }
        require(i <= limit) { INVALID_STRING }
        return String(buf, 0, j)
    }

    private fun encodeStr(s: String) {
        val buf = hb
        var j = position
        var i = 0
        val n = s.length
        while (i < n) {
            val c = s[i++]
            if (c.code < 0x80) {
                // 0xxxxxxx
                buf[j++] = c.code.toByte()
            } else if (c.code < 0x800) {
                // 110xxxxx 10xxxxxx
                buf[j++] = (0xC0 or (c.code ushr 6)).toByte()
                buf[j++] = (0x80 or (0x3F and c.code)).toByte()
            } else if (c.code < 0xD800 || c.code > 0xDFFF) {
                // 1110xxxx 10xxxxxx 10xxxxxx
                buf[j++] = (0xE0 or (c.code ushr 12)).toByte()
                buf[j++] = (0x80 or (0x3F and (c.code ushr 6))).toByte()
                buf[j++] = (0x80 or (0x3F and c.code)).toByte()
            } else {
                // 11110xxx 10xxxxxx 10xxxxxx 10xxxxxx
                val c2 = s[i++]
                val cp = (c.code shl 10) + c2.code + -0x35fdc00
                buf[j++] = (0xF0 or (cp ushr 18)).toByte()
                buf[j++] = (0x80 or (0x3F and (cp ushr 12))).toByte()
                buf[j++] = (0x80 or (0x3F and (cp ushr 6))).toByte()
                buf[j++] = (0x80 or (0x3F and cp)).toByte()
            }
        }
        position = j
    }

    fun getChecksum(start: Int, size: Int): Long {
        if (size <= 0) return 0L
        var p = start
        val n = size shr 3
        val remain = size and 7
        var checkSum = 0L
        for (i in 0 until n) {
            checkSum = checkSum xor getLong(p)
            p += 8
        }
        val maxShift = remain shl 3
        var i = 0
        while (i < maxShift) {
            checkSum = checkSum xor (hb[p++].toLong() and 0xFFL shl i)
            i += 8
        }
        val shift = start and 7 shl 3
        return checkSum shl shift or (checkSum ushr 64 - shift)
    }

    companion object {
        private const val MAX_CHAR_LEN = 2048
        private const val INVALID_STRING = "Invalid String"

        @JvmStatic
        fun getVarint32Size(v: Int): Int {
            if (v shr 7 == 0) {
                return 1
            } else if (v shr 14 == 0) {
                return 2
            } else if (v shr 21 == 0) {
                return 3
            } else if (v shr 28 == 0) {
                return 4
            }
            return 5
        }

        /**
         * get string size in utf-8
         *
         * @param str The String
         * @return The utf-8 size of the string
         */
        @JvmStatic
        fun getStringSize(str: String): Int {
            var j = 0
            var i = 0
            val n = str.length
            while (i < n) {
                val c = str[i++]
                j +=
                    if (c.code < 0x80) {
                        1
                    } else if (c.code < 0x800) {
                        2
                    } else if (c.code < 0xD800 || c.code > 0xDFFF) {
                        3
                    } else {
                        i++
                        4
                    }
            }
            return j
        }
    }
}
