/*
 * Copyright (c) 田梓萱[小草林] 2021-2024.
 * All Rights Reserved.
 * All codes are protected by China's regulations on the protection of computer software, and infringement must be investigated.
 * 版权所有 (c) 田梓萱[小草林] 2021-2024.
 * 所有代码均受中国《计算机软件保护条例》保护，侵权必究.
 */
package io.fastkv

import android.annotation.SuppressLint
import java.io.Closeable
import java.io.File
import java.io.IOException
import java.io.RandomAccessFile
import java.security.SecureRandom
import java.util.concurrent.atomic.AtomicBoolean

internal object Utils {
    const val NAME_SIZE = 32

    /** Like UUID.randomUUID().toString(), but without '-'. */
    @JvmStatic
    fun randomName(): String {
        val len = NAME_SIZE shr 1
        val bytes = ByteArray(len)
        Holder.random.nextBytes(bytes)
        val buf = CharArray(NAME_SIZE)
        for (i in 0 until len) {
            val b = bytes[i].toInt()
            val index = i shl 1
            buf[index] = Holder.digits[b shr 4 and 0xF]
            buf[index + 1] = Holder.digits[b and 0xF]
        }
        return String(buf)
    }

    @JvmStatic
    @get:SuppressLint("DiscouragedPrivateApi")
    val pageSize: Int
        get() {
            try {
                val unsafeClass = Class.forName("sun.misc.Unsafe")
                val theUnsafe = unsafeClass.getDeclaredField("theUnsafe")
                theUnsafe.isAccessible = true
                val method = unsafeClass.getDeclaredMethod("pageSize")
                method.isAccessible = true
                return method.invoke(theUnsafe[null]) as Int
            } catch (ignore: Throwable) {}
            return 4096
        }

    @JvmStatic
    @Throws(IOException::class)
    fun makeFileIfNotExist(file: File): Boolean {
        return if (file.isFile) {
            true
        } else {
            val parent = file.parentFile
            parent != null && (parent.isDirectory || parent.mkdirs()) && file.createNewFile()
        }
    }

    @JvmStatic
    @Throws(IOException::class)
    fun getBytes(file: File): ByteArray? {
        if (!file.isFile) {
            return null
        }
        val len = file.length()
        require(len shr 32 == 0L) { "file too large, path:" + file.path }
        val bytes = ByteArray(len.toInt())
        readBytes(file, bytes, len.toInt())
        return bytes
    }

    @JvmStatic
    @Throws(IOException::class)
    fun readBytes(file: File?, bytes: ByteArray?, len: Int) {
        val accessFile = RandomAccessFile(file, "rw")
        try {
            var p = 0
            while (p < len) {
                val count = accessFile.read(bytes, p, len - p)
                if (count < 0) break
                p += count
            }
        } finally {
            closeQuietly(accessFile)
        }
    }

    fun saveBytes(dstFile: File, bytes: ByteArray, canceled: AtomicBoolean?): Boolean {
        var tmpFile: File? = null
        try {
            val len = bytes.size
            tmpFile = File(dstFile.getParent(), dstFile.getName() + ".tmp")
            if (!makeFileIfNotExist(tmpFile)) {
                logError(java.lang.Exception("create file failed"))
                return false
            }
            if (canceled != null && canceled.get()) {
                return false
            }
            RandomAccessFile(tmpFile, "rw").use { accessFile ->
                accessFile.setLength(len.toLong())
                accessFile.write(bytes, 0, len)
                if (canceled != null && canceled.get()) {
                    return false
                }
                accessFile.getFD().sync()
            }
            return renameFile(tmpFile, dstFile)
        } catch (e: java.lang.Exception) {
            logError(java.lang.Exception("save bytes failed", e))
        } finally {
            if (canceled != null && canceled.get()) {
                deleteFile(tmpFile!!)
                deleteFile(dstFile)
            }
        }
        return false
    }

    @JvmStatic
    fun renameFile(srcFile: File, dstFile: File): Boolean {
        return if (srcFile.renameTo(dstFile)) {
            true
        } else (!dstFile.exists() || dstFile.delete()) && srcFile.renameTo(dstFile)
    }

    @JvmStatic
    fun closeQuietly(closeable: Closeable?) {
        if (closeable != null) {
            try {
                closeable.close()
            } catch (ignore: Throwable) {}
        }
    }

    @JvmStatic
    fun deleteFile(file: File) {
        try {
            if (file.exists()) {
                deleteRecursive(file)
            }
        } catch (ignore: Throwable) {}
    }

    private fun deleteRecursive(file: File) {
        if (file.isDirectory) {
            val fs = file.listFiles()
            if (fs != null) {
                for (f in fs) {
                    deleteRecursive(f)
                }
            }
        }
        file.delete()
    }

    @JvmStatic
    fun moveDirFiles(srcDir: File, currentDir: String?) {
        if (srcDir.isDirectory) {
            val fs = srcDir.listFiles()
            if (fs != null) {
                for (file in fs) {
                    try {
                        moveFile(file, File(currentDir, file.name))
                    } catch (e: Exception) {
                        logError(e)
                    }
                }
            }
        }
    }

    @Throws(IOException::class)
    private fun moveFile(srcFile: File, dstFile: File) {
        if (!srcFile.exists() || dstFile.exists()) {
            return
        }
        if (!srcFile.renameTo(dstFile)) {
            getBytes(srcFile)?.let { saveBytes(dstFile, it, null) }
            deleteFile(srcFile)
        }
    }

    @JvmStatic
    fun binarySearch(a: IntArray, value: Int): Int {
        var lo = 0
        var hi = a.size - 1
        while (lo <= hi) {
            val mid = lo + hi ushr 1
            val midVal = a[mid]
            if (midVal < value) {
                lo = mid + 1
            } else if (midVal > value) {
                hi = mid - 1
            } else {
                return mid
            }
        }
        return hi
    }

    private fun logError(e: Exception) {
        val logger = FastKVConfig.sLogger
        logger.e("FastKV", e)
    }

    private object Holder {
        val random = SecureRandom()
        val digits =
            charArrayOf(
                '0', '1', '2', '3', '4', '5', '6', '7', '8', '9', 'a', 'b', 'c', 'd', 'e', 'f')
    }
}
